###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import colorsys
from string import Template

from vistrails.core.modules.config import IPort, OPort, ModuleSettings
from vistrails.core.modules.vistrails_module import ModuleError, Module
from vistrails.core.utils import InstanceObject
# FIXME make package imports better
from vistrails.packages.tabledata.common import choose_column

from .utils import *

###############################################################################
# Modules and Data

class GMapVisData(object):
    def __init__(self, gmaps_libs, init_template, data, center=None):
        self.gmaps_libs = gmaps_libs
        self.init_template = init_template
        self.data = data
        self.center = center

def convert_color(c):
    c = c.tuple
    return GMapColor(c[0] * 255.0, c[1] * 255.0, c[2] * 255.0)

class TitlesMixin(object):
    def get_titles(self, table=None, default_col=None):
        if table is None:
            table = self.get_input("table")
        title_col_idx = self.force_get_input('titleColIdx')
        title_col_name = self.force_get_input('titleColName')
        print "title_col_idx:", title_col_idx
        print "title_col_name:", title_col_name

        if (title_col_idx is None and
                title_col_name is None and 
                default_col is not None and
                default_col < table.columns):
                # only allow default if in range
                title_col_idx = default_col

        if title_col_idx is not None or title_col_name is not None:
            title_idx = choose_column(table.columns, table.names, 
                                      title_col_name, title_col_idx)
            title_col = table.get_column(title_idx)
            return title_col
        return None

class GMapVis(Module, OptionsMixin):
    _input_ports = [IPort('table', 'tabledata:Table'),
                    IPort('latitudeColIdx', 'basic:Integer', optional=True,
                          default=0),
                    IPort('latitudeColName', 'basic:String', optional=True),
                    IPort('longitudeColIdx', 'basic:Integer', optional=True,
                          default=1),
                    IPort('longitudeColName', 'basic:String', optional=True)]
    _output_ports = [OPort('self', 'GMapVis')]
    _settings = ModuleSettings(abstract=True)

    def get_positions(self, table=None):
        if table is None:
            table = self.get_input("table")
        
        lat_col_idx = self.force_get_input('latitudeColIdx')
        lat_col_name = self.force_get_input('latitudeColName')
        lng_col_idx = self.force_get_input('longitudeColIdx')
        lng_col_name = self.force_get_input('longitudeColName')
        if lat_col_idx is None and lat_col_name is None:
            lat_idx = self.get_input('latitudeColIdx') # default 0
        else:
            lat_idx = choose_column(table.columns, table.names, 
                                    lat_col_name, lat_col_idx)
        if lng_col_idx is None and lng_col_name is None:
            lng_idx = self.get_input('longitudeColIdx') # default 1
        else:
            lng_idx = choose_column(table.columns, table.names, 
                                    lng_col_name, lng_col_idx)
        lat_col = table.get_column(lat_idx, True)
        lng_col = table.get_column(lng_idx, True)
        center = (sum(float(x) for x in lat_col)/len(lat_col),
                  sum(float(x) for x in lng_col)/len(lng_col))
        positions = []
        for i in xrange(table.rows):
            positions.append(GMapLatLng(lat_col[i], lng_col[i]))
        return (positions, center)

class GMapMarkers(GMapVis, TitlesMixin):
    TEMPLATE = Template("""
  var positions = $marker_data;
  var options = $marker_options;
  var titles = $marker_titles;

  for (var i=0; i < positions.length; i++) {
    marker = new google.maps.Marker({"position": positions[i],
                                      "map": map});
    marker.setOptions(options);
    if (titles) {
      marker.setTitle(titles[i]);
    }
  }
""")
    SPECS = ['flat']
    _input_ports = [IPort("flat", "basic:Boolean", optional=True),
                    IPort('titleColIdx', 'basic:Integer', optional=True),
                    IPort('titleColName', 'basic:String', optional=True)]
    
    def compute(self):
        (positions, center) = self.get_positions()
        marker_options = self.get_options(self.SPECS)
        titles = self.get_titles()
        print "got titles:", titles
        data = {"marker_options": marker_options,
                "marker_data": positions,
                "marker_titles": titles}
        
        vis_data = GMapVisData([], self.TEMPLATE, data, center)
        self.set_output("self", vis_data)

class GMapValueVis(GMapVis):
    _input_ports = [IPort('valueColIdx', 'basic:Integer', optional=True,
                          default=2),
                    IPort('valueColName', 'basic:String', optional=True),]
    _settings = ModuleSettings(abstract=True)
    def get_values(self, table=None):
        if table is None:
            table = self.get_input("table")
        value_col_idx = self.force_get_input("valueColIdx")
        value_col_name = self.force_get_input("valueColName")
        if value_col_idx is None and value_col_name is None:
            value_idx = self.get_input("valueColIdx") # default 2
        else:
            value_idx = choose_column(table.columns, table.names, 
                                      value_col_name, value_col_idx)
        value_col = table.get_column(value_idx, True)
        return value_col

class GMapCircles(GMapValueVis):
    TEMPLATE = Template("""
  var data = $circle_data;
  
  for (var i=0; i < data.length; i++) {
    var options = $circle_options;
    options["center"] = data[i][0];
    options["radius"] = data[i][1];
    options["map"] = map;
    circle = new google.maps.Circle(options);
  }
""")
    
    SPECS = [('strokeColor', convert_color, True),
             ('fillColor', convert_color),
             'strokeWeight', 
             'strokeOpacity',
             'fillOpacity']
    _input_ports = [IPort("strokeColor", "basic:Color", optional=True,
                          default=InstanceObject(tuple=(0,0,0))),
                    IPort("strokeWeight", "basic:Integer", optional=True),
                    IPort("strokeOpacity", "basic:Float", optional=True),
                    IPort("fillColor", "basic:Color", optional=True),
                    IPort("fillOpacity", "basic:Float", optional=True),
                    IPort("scale", "basic:Float", optional=True)]
    
    def compute(self):
        (positions, center) = self.get_positions()
        values = self.get_values()
        circle_data = [[positions[i], float(values[i])/200.0]
                       for i in xrange(len(positions))]
        circle_options = self.get_options(self.SPECS)
        data = {"circle_options": circle_options,
                "circle_data": circle_data}
        vis_data = GMapVisData([], self.TEMPLATE, data, center)
        self.set_output("self", vis_data)

class GMapSymbols(GMapValueVis, TitlesMixin):
    TEMPLATE = Template("""
  var data = $symbol_data;
  var titles = $symbol_titles;

  for (var i=0; i < data.length; i++) {
    var marker_options = {"position": data[i][0],
                          "map": map};
    if (titles) {
      marker_options["title"] = titles[i];
    }
    if ($use_values) {
      var icon_options = $symbol_options;
      icon_options["fillColor"] = data[i][1];
      marker_options["icon"] = icon_options;
    }
    marker = new google.maps.Marker(marker_options);
  }  
""")

    SPECS = [('strokeColor', convert_color, True),
             ('fillStartColor', None, True),
             ('fillEndColor', None, True),
             ('strokeWeight', None, True),
             'strokeOpacity',
             ('fillOpacity', None, True),
             ('scale', None, True)]
    _input_ports = [IPort("strokeColor", "basic:Color", optional=True,
                          default=InstanceObject(tuple=(0,0,0))),
                    IPort("strokeWeight", "basic:Integer", optional=True,
                          default=1),
                    IPort("strokeOpacity", "basic:Float", optional=True),
                    IPort("fillStartColor", "basic:Color", optional=True,
                          default=InstanceObject(tuple=(1,1,1))),
                    IPort("fillEndColor", "basic:Color", optional=True,
                          default=InstanceObject(tuple=(1,0,0))),
                    IPort("fillOpacity", "basic:Float", optional=True,
                          default=1.0),
                    IPort("scale", "basic:Float", optional=True,
                          default=5.0),
                    IPort('titleColIdx', 'basic:Integer', optional=True),
                    IPort('titleColName', 'basic:String', optional=True),
                    IPort("allowLegacy", "basic:Boolean", optional=True,
                          default=False)]

    def compute(self):
        (positions, center) = self.get_positions()
        legacy = self.get_input("allowLegacy")
        use_values = True
        try:
            values = [float(x) for x in self.get_values()]
        except ValueError, e:
            # LEGACY SUPPORT
            if legacy:
                use_values = False
            else:
                raise ModuleError(self, "Must provide values column")
        if not use_values:
            symbol_data = positions
            symbol_options = {}
        else:
            symbol_options = self.get_options(self.SPECS)
            symbol_options["path"] = \
                            RawJavaScriptText("google.maps.SymbolPath.CIRCLE")
            min_value = min(values)
            max_value = max(values)

            # if we have black or white, we want hue to match the other side
            def white_or_black(c):
                return ((c[0] < 1e-8 and c[1] < 1e-8 and c[2] < 1e-8) or
                        (c[0] > 1-1e-8 and c[1] > 1-1e-8 and c[2] > 1-1e-8))
            start_c = symbol_options.pop("fillStartColor").tuple
            end_c = symbol_options.pop("fillEndColor").tuple
            start_wb = white_or_black(start_c)
            end_wb = white_or_black(end_c)
            start_c = list(colorsys.rgb_to_hsv(*start_c))
            end_c = list(colorsys.rgb_to_hsv(*end_c))
            if start_wb:
                start_c[0] = end_c[0]
            elif end_wb:
                end_c[0] = start_c[0]

            symbol_data = []
            for i in xrange(len(positions)):
                val = values[i]
                if max_value - min_value < 1e-8:
                    norm_val = 1.0
                else:
                    norm_val = (val - min_value) / (max_value - min_value)
                color = []
                for j in xrange(len(start_c)):
                    color.append((1.0 - norm_val) * start_c[j] + 
                                 norm_val * end_c[j])
                color = colorsys.hsv_to_rgb(*color)
                symbol_data.append([positions[i],
                                    GMapColor(255 * color[0],
                                              255 * color[1],
                                              255 * color[2])])
        symbol_titles = self.get_titles(default_col=(3 if legacy else None))
        data = {"symbol_data": symbol_data,
                "symbol_options": symbol_options,
                "symbol_titles": symbol_titles,
                "use_values": use_values}
        vis_data = GMapVisData([], self.TEMPLATE, data, center)
        self.set_output("self", vis_data)

class GMapHeatmap(GMapValueVis):
    TEMPLATE = Template("""
  var data = $heatmap_data;
  var options = $heatmap_options;
  options["data"] = data;
  options["map"] = map;
  heatmap = new google.maps.visualization.HeatmapLayer(options);
""")

    SPECS = ['dissipating',
             'maxIntensity',
             'opacity',
             'radius']
    _input_ports = [IPort("dissipating", "basic:Boolean", optional=True,
                          default=True),
                    IPort("maxIntensity", "basic:Float", optional=True),
                    IPort("opacity", "basic:Float", optional=True,
                     default=0.6),
                    IPort("radius", "basic:Float", optional=True)]
    
    def compute(self):
        (positions, center) = self.get_positions()
        values = self.get_values()
        heatmap_data = [{"location": positions[i],
                         "weight": float(values[i])} 
                        for i in xrange(len(positions))]
        heatmap_options = self.get_options(self.SPECS)
        data = {"heatmap_data": heatmap_data,
                "heatmap_options": heatmap_options}
        vis_data = GMapVisData([], self.TEMPLATE, data, center)
        self.set_output("self", vis_data)

_modules = [GMapVis, GMapMarkers, GMapValueVis, GMapCircles, GMapSymbols,
            GMapHeatmap]



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains a dialog and widgets related to the module annotation
displaying a list of all pairs (key,value) for a module

QKeyValueDelegate
QModuleAnnotation
QModuleAnnotationTable
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core import debug

################################################################################

class QModuleAnnotation(QtGui.QDialog):
    """
    QModuleAnnotation is a dialog for annotating modules

    """
    def __init__(self, module, controller, parent=None):
        """ 
        QModuleAnnotation(module: Module, controller: VistrailController)
        -> None

        """
        QtGui.QDialog.__init__(self, parent)
        self.module = module
        self.controller = controller
        self.setModal(True)
        self.setWindowTitle('Module Annotations')
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().setMargin(0)
        self.layout().setSpacing(0)
        self.scrollArea = QtGui.QScrollArea(self)
        self.layout().addWidget(self.scrollArea)
        self.scrollArea.setFrameStyle(QtGui.QFrame.NoFrame)
        self.annotationTable = QModuleAnnotationTable(self.module,
                                                      self.controller,
                                                      self)
        self.scrollArea.setWidget(self.annotationTable)
        self.scrollArea.setWidgetResizable(True)
        self.buttonLayout = QtGui.QHBoxLayout()
        self.buttonLayout.setMargin(5)
        self.closeButton = QtGui.QPushButton('Close', self)
        self.closeButton.setFixedWidth(100)
        self.buttonLayout.addWidget(self.closeButton)
        self.closeButton.setShortcut('Esc')
        self.layout().addLayout(self.buttonLayout)
        self.connect(self.closeButton, QtCore.SIGNAL('clicked(bool)'), self.close)

        
class QModuleAnnotationTable(QtGui.QTableWidget):
    """
    QModuleAnnotationTable is a table widget that can be dock inside a
    window. It has two columns for key and value pairs to view/edit at
    run-time
    
    """    
    def __init__(self, module=None, controller=None, parent=None):
        """ QModuleAnnotationTable(module: Module, controller: 
        VistrailController, parent: QWidget) -> QModuleAnnotationTable
        Construct the 1x2 table
        
        """
        QtGui.QTableWidget.__init__(self, 1, 2, parent)
        self.read_only = False
        self.setHorizontalHeaderLabels(['Key', 'Value'])
        self.horizontalHeader().setResizeMode(QtGui.QHeaderView.Interactive)
        self.horizontalHeader().setMovable(False)
        self.horizontalHeader().setStretchLastSection(True)
        self.setSortingEnabled(True)
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        self.verticalHeader().hide()
        self.delegate = QKeyValueDelegate(self)
        self.setItemDelegate(self.delegate)
        self.module = module
        self.controller = controller
        self.updateLocked = False
        self.updateModule()

    def set_controller(self, controller):
        self.controller = controller

    def setReadOnly(self, read_only):
        if read_only != self.read_only:
            self.read_only = read_only
            self.setEnabled(not read_only and self.module is not None)

    def updateModule(self, module=None):
        """ updateModule() -> None
        Update the widget to view the module annotations
        """

        self.module = module
        self.setSortingEnabled(False)
        if self.updateLocked: return
        self.clearContents()
        self.setRowCount(0)
        if self.module:
            self.setRowCount(len(self.module.annotations)+1)
            curRow = 0
            for annotation in self.module.annotations:
                if annotation.key == '__desc__':
                    # We don't display the '__desc__' annotation in the list
                    # anymore. If it's present we decrease the rowcount by 1
                    self.setRowCount(len(self.module.annotations))
                else:
                    self.setItem(curRow, 0, QtGui.QTableWidgetItem(annotation.key))
                    item = QtGui.QTableWidgetItem(annotation.value)
                    self.setItem(curRow, 1, item)
                    curRow += 1
            self.setEnabled(not self.read_only)
        else:
            self.setRowCount(1)
            self.setEnabled(False)
        self.setItem(self.rowCount()-1, 0, QtGui.QTableWidgetItem(''))
        self.setItem(self.rowCount()-1, 1, QtGui.QTableWidgetItem(''))
        self.setSortingEnabled(True)

    def makeItemBold(self, index):
        """ makeItemBold(index: QModelIndex) -> None
        Make the item at index to have a bold face
        
        """
        oldFont = QtGui.QFont(self.model().data(index, QtCore.Qt.FontRole))
        oldFont.setBold(True)
        oldFont.setPointSize(20)
        self.model().setData(index, oldFont, QtCore.Qt.FontRole)

    def lockUpdate(self):
        """ lockUpdate() -> None
        Do not allow updateModule()
        
        """
        self.updateLocked = True
        
    def unlockUpdate(self):
        """ unlockUpdate() -> None
        Allow updateModule()
        
        """
        self.updateLocked = False

    def addRow(self):
        """ addRow() -> None
        Adds a new empty row to the table

        """
        self.setSortingEnabled(False)
        self.resizeRowsToContents()
        self.insertRow(self.rowCount())
        self.setItem(self.rowCount()-1, 0,
                     QtGui.QTableWidgetItem(''))
        self.setItem(self.rowCount()-1, 1,
                     QtGui.QTableWidgetItem(''))
        self.setSortingEnabled(False)

    def editNextAvailableCell(self):
        item = self.item(self.rowCount()-1, 0)
        self.editItem(item)
        
class QKeyValueDelegate(QtGui.QItemDelegate):
    """    
    QKeyValueDelegate tries to create a special control widget
    providing a simple interface for adding/deleting module
    annotations
    
    """

    def __init__(self, table):
        """ QKeyValueDelegate(table: QModuleAnnotationTable) -> QKeyValueDelegate
        Save a reference to table and perform a default initialization
        
        """
        self.table = table
        QtGui.QItemDelegate.__init__(self, None)
    
    def setEditorData(self, editor, index):
        """ setEditorData(editor: QWidget, index: QModelIndex) -> None
        Set the current item (at index) data into editor for editting
        
        """
        text = index.data(QtCore.Qt.DisplayRole)
        editor.setText(text)

    def setModelData(self, editor, model, index):
        """ setModelData(editor: QWidget, model: QAbstractItemModel,
                         index: QModelIndex) -> None                         
        Assign the value of the editor back into the model and emit a
        signal to update vistrail
        
        """
        text = str(editor.text())
        row = index.row()
        col = index.column()
        keyItem = self.table.item(row, 0)
        if keyItem:
            key = str(keyItem.text())
        else:
            key = ''
            
        valueItem = self.table.item(row, 1)
        if valueItem:
            value = str(valueItem.text())
        else:
            value = ''
            
        if col==0:
            if text=='' and row<self.table.rowCount()-1:
                self.table.removeRow(row)
                if self.table.controller and self.table.module:
                    self.table.lockUpdate()
                    self.table.controller.delete_annotation(key,
                                                            self.table.module.id)
                    self.table.unlockUpdate()
                return
            if text!='' and text!=key:
                if (self.table.module and
                    self.table.module.has_annotation_with_key(text)):
                    if text == '__desc__':
                        QtGui.QMessageBox.information(None,
                                                      "VisTrails",
                                    'Please use "Set Module Label..." menu option'
                                    ' to set the label for this module.')
                    else:
                        QtGui.QMessageBox.information(None,
                                                  "VisTrails",
                                                  text + ' already exists in '
                                                  'the annotations.')
                    return

        if col==1 and key=='':
            QtGui.QMessageBox.information(None,
                                          "VisTrails",
                                          "Must provide a key first.")
            return
            
            
        if col==0 and key=='' and text!='' and value!='':
            self.table.addRow()
                    
        if col==1:
            if text!=value:
                if self.table.controller and self.table.module:
                    self.table.lockUpdate()
                    self.table.controller.add_annotation((key, text),
                                                         self.table.module.id)
                    self.table.unlockUpdate()
            if row == self.table.rowCount()-1:
                self.table.addRow()
            
        elif text!='' and self.table.controller and self.table.module:
            moduleId = self.table.module.id
            self.table.lockUpdate()
            self.table.controller.previousModuleIds = [moduleId]
            if key!=text and key!='':
                self.table.controller.delete_annotation(key, moduleId)
            self.table.controller.add_annotation((text, value),
                                                 moduleId)
            self.table.unlockUpdate()
        
        model.setData(index, text)        



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" Mac OS X specific file """

# from xml import dom
# from xml.dom.xmlbuilder import DOMInputSource, DOMBuilder
from __future__ import division

import xml.etree.cElementTree as ElementTree
import datetime
import os
import shutil
import subprocess
import tempfile

from vistrails.core.system import time_strptime
from vistrails.core.system.unix import executable_is_in_path, list2cmdline, \
     execute_cmdline, execute_piped_cmdlines, execute_cmdline2
from vistrails.core import debug
import vistrails.core.utils


__all__ = ['executable_is_in_path', 'list2cmdline',
           'execute_cmdline', 'execute_piped_cmdlines', 'execute_cmdline2',

           'guess_total_memory',
           'home_directory', 'remote_copy_program', 'remote_shell_program',
           'graph_viz_dot_command_line', 'remove_graph_viz_temporaries',
           'link_or_copy', 'get_executable_path',
           'shell_font_face', 'shell_font_size',
           'TestMacOSX']

###############################################################################
# Extract system detailed information of a Mac system
#
# Based on a Python Cookbook recipe available online :
#    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/303063
# and in the Python Coobook, page 418
# Credit: Brian Quinlan
#
# Rewritten to use ElementTree instead of PyXML
#
# This recipe uses the system_profiler application to retrieve detailed
# information about a Mac OS X system. There are two useful ways to use it:
# the first is to ask for a complete Python datastructure containing
# information about the system (see OSXSystemProfiler.all()) and the
# other is two ask for particular keys in the system information database.

def group(lst, n):
    """group([0,3,4,10,2,3], 2) => [(0,3), (4,10), (2,3)]

    Group a list into consecutive n-tuples. Incomplete tuples are
    discarded e.g.

    >>> group(range(10), 3)
    [(0, 1, 2), (3, 4, 5), (6, 7, 8)]

    """
    return zip(*[lst[i::n] for i in xrange(n)])

class OSXSystemProfiler(object):
    "Provide information from the Mac OS X System Profiler"

    def __init__(self, category=None, detail=None):
        """detail can range from -2 to +1, with larger numbers returning more
        information. Beware of +1, it can take several minutes for
        system_profiler to generate the data."""

        command = ['system_profiler', '-xml']
        if category is not None:
            command.append(str(category))
        if detail is not None:
            command.extend(['-detailLevel', '%d' % detail])
        p = subprocess.Popen(command, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        stderr = stderr.strip()
        if stderr:
            lines = stderr.splitlines()
            if len(lines) > 1 or len(lines[0]) > 44:
                line = "%s..." % lines[0][:41]
            else:
                line = lines[0]
            debug.warning("Error output from system_profiler: %s" % line,
                          stderr)
        self.document = ElementTree.XML(stdout)

    def _content(self, node):
        "Get the text node content of an element or an empty string"
        return (node.text or '')

    def _convert_value_node(self, node):
        """Convert a 'value' node (i.e. anything but 'key') into a Python data
        structure"""
        if node.tag == 'string':
            return self._content(node)
        elif node.tag == 'integer':
            return int(self._content(node))
        elif node.tag == 'real':
            return float(self._content(node))
        elif node.tag == 'date': #  <date>2004-07-05T13:29:29Z</date>
            return datetime.datetime(
                *time_strptime(self._content(node), '%Y-%m-%dT%H:%M:%SZ')[:5])
        elif node.tag == 'array':
            return [self._convert_value_node(n) for n in node.getchildren()]
        elif node.tag == 'dict':
            return dict([(self._content(n), self._convert_value_node(m))
                for n, m in group(node.getchildren(), 2)])
        else:
            raise ValueError(node.tag)

    def __getitem__(self, key):
        nodes = self.document.getiterator('dict')
        results = []
        for node in nodes:
            for child in node:
                if child.tag == 'key' and child.text == key:
                    v = self._convert_value_node(node)[key]
                    if isinstance(v, dict) and v.has_key('_order'):
                        # this is just information for display
                        pass
                    else:
                        results.append(v)
        return results

###############################################################################

def parse_meminfo():
    """ parse_meminfo() -> int
    Uses the system_profiler application to retrieve detailed information
    about a Mac OS X system. Returns memory size in Megabytes.

    Just use the "SPHardwareDataType" category to limit the amount of
    information gathered.

    """

    result = -1
    info = OSXSystemProfiler("SPHardwareDataType")
    mem = info['physical_memory'][0]
    # print "*** MEMORY", mem
    if mem.upper().endswith(' GB'):
        result = int(float(mem[:-3]) * 1024) * 1L
    elif mem.upper().endswidth(' MB'):
        result = int(mem[:-3]) * 1L
    # print '>>>>', result
    return result

def guess_total_memory():
    """ guess_total_memory() -> int
    Return system memory in bytes. If PyXML is not installed it returns -1

    """
    return parse_meminfo()

def home_directory():
    """ home_directory() -> str
    Returns user's home directory using environment variable $HOME

    """
    return os.getenv('HOME')

def remote_copy_program():
    return "scp -p"

def remote_shell_program():
    return "ssh -p"

def graph_viz_dot_command_line():
    """ graph_viz_dot_command_line() -> str
    Returns dot command line

    """
    return 'dot -Tplain -o '

def remove_graph_viz_temporaries():
    """ remove_graph_viz_temporaries() -> None
    Removes temporary files generated by dot

    """
    os.unlink(tempfile.gettempdir() + "dot_output_vistrails.txt")
    os.unlink(tempfile.gettempdir() + "dot_tmp_vistrails.txt")

def link_or_copy(src, dst):
    """link_or_copy(src:str, dst:str) -> None
    Tries to create a hard link to a file. If it is not possible, it will
    copy file src to dst

    """
    # Links if possible, but we're across devices, we need to copy.
    try:
        os.link(src, dst)
    except OSError, e:
        if e.errno == 18:
            # Across-device linking is not possible. Let's copy.
            shutil.copyfile(src, dst)
        else:
            raise e

def get_executable_path(executable_name):
    """get_executable_path(executable_name: str) -> str
    Get the absolute filename of an executable, searching in the PATH.
    """
    pathlist = os.environ['PATH'].split(os.pathsep)
    vt_path = os.getenv("EXECUTABLEPATH")
    if vt_path:
        pathlist.insert(0, vt_path.strip())
    for path in pathlist:
        fullpath = os.path.join(path, executable_name)
        if os.path.isfile(fullpath):
            return os.path.abspath(fullpath)

def shell_font_face():
    return 'Monaco'

def shell_font_size():
    return 12

################################################################################

import unittest

class TestMacOSX(unittest.TestCase):
    """ Class to test Mac OS X specific functions """

    def test1(self):
        """ Test if guess_total_memory() is returning an int >= 0"""
        result = guess_total_memory()
        assert isinstance(result, (int, long))
        assert result >= 0

    def test2(self):
        """ Test if home_directory is not empty """
        result = home_directory()
        assert result != ""

    def test_executable_file_in_path(self):
        # Should exist in any POSIX shell, which is what we have in OSX
        self.assertTrue(executable_is_in_path('ls'))

if __name__ == '__main__':
    unittest.main()




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

if __name__ == '__main__':
    import sip
    sip.setapi('QString', 2)
    sip.setapi('QVariant', 2)

import json
import os
import platform
import string
import subprocess
import sys
import threading
from PyQt4 import QtCore, QtGui

encode_list = [['\xe2\x80\x90', '-'],
               ['\xe2\x80\x9d', '"'],
               ['\xe2\x80\x9c', '"']]

SUFFIX = '.clt'

def default_dir():
    systemType = platform.system()
    if systemType in ['Windows', 'Microsoft']:
        if len(os.environ['HOMEPATH']) == 0:
            root = '\\'
        else:
            root = os.environ['HOMEDRIVE'] + os.environ['HOMEPATH']
    else:
        root = os.getenv('HOME')
    default_dir = os.path.join(root, '.vistrails', 'CLTools')
    if not os.path.exists(default_dir):
        os.mkdir(default_dir)
    return default_dir

def quote_arg(arg):
    arg = arg.replace('\\', '\\\\')
    if '"' in arg or any(c in arg for c in string.whitespace):
        return '"%s"' % arg.replace('"', '\\"')
    else:
        return arg

# From: https://gist.github.com/kirpit/1306188
class Command(object):
    """
    Enables to run subprocess commands in a different thread with TIMEOUT option.
 
    Based on jcollado's solution:
    http://stackoverflow.com/questions/1191374/subprocess-with-timeout/4825933#4825933
    """
    command = None
    process = None
    status = None
    output, error = '', ''
 
    def __init__(self, command):
        self.command = command
 
    def run(self, timeout=5, **kwargs):
        """ Run a command then return: (status, output, error). """
        def target(**kwargs):
            try:
                self.process = subprocess.Popen(self.command, **kwargs)
                self.output, self.error = self.process.communicate()
                self.status = self.process.returncode
            except Exception:
                import traceback
                self.error = traceback.format_exc()
                self.status = -1
        # default stdout and stderr
        if 'stdout' not in kwargs:
            kwargs['stdout'] = subprocess.PIPE
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.PIPE
        # thread
        print "calling with kwargs", target, kwargs
        thread = threading.Thread(target=target, kwargs=kwargs)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            self.process.terminate()
            thread.join()
        return self.status, self.output, self.error

class QCLToolsWizard(QtGui.QWidget):
    def __init__(self, parent, reload_scripts=None):
        QtGui.QWidget.__init__(self, parent)

        self.vbox = QtGui.QVBoxLayout()
        self.vbox.setContentsMargins(5,5,5,5)

        self.setLayout(self.vbox)
        self.setTitle()
        self.file = None
        self.conf = None
        self.reload_scripts = reload_scripts

        self.toolBar = QtGui.QToolBar()
        self.layout().addWidget(self.toolBar)
        self.newFileAction = QtGui.QAction(
            self.get_icon('document-new'), 'New', self)
        self.newFileAction.setToolTip('Start on a new Wrapper')
        self.connect(self.newFileAction, QtCore.SIGNAL('triggered()'),
                     self.newFile)
        self.toolBar.addAction(self.newFileAction)
        self.openFileAction = QtGui.QAction(
            self.get_icon('document-open'), 'Open', self)
        self.openFileAction.setToolTip('Open an existing wrapper')
        self.connect(self.openFileAction, QtCore.SIGNAL('triggered()'),
                     self.openFile)
        self.toolBar.addAction(self.openFileAction)
        self.saveFileAction = QtGui.QAction(
            self.get_icon('document-save'), 'Save', self)
        self.saveFileAction.setToolTip('Save wrapper')
        self.connect(self.saveFileAction, QtCore.SIGNAL('triggered()'),
                     self.save)
        self.toolBar.addAction(self.saveFileAction)
        self.saveFileAsAction = QtGui.QAction(
            self.get_icon('document-save-as'), 'Save As', self)
        self.saveFileAsAction.setToolTip('Save wrapper as a new file')
        self.connect(self.saveFileAsAction, QtCore.SIGNAL('triggered()'),
                     self.saveAs)
        self.toolBar.addAction(self.saveFileAsAction)

        if self.reload_scripts:
            self.reloadAction = QtGui.QAction(
                self.get_icon('view-refresh'), 'Refresh', self)
            self.reloadAction.setToolTip('Save and Reload CLTools Modules in VisTrails')
            self.connect(self.reloadAction, QtCore.SIGNAL('triggered()'),
                         self.refresh)
            self.toolBar.addAction(self.reloadAction)
        
        self.toolBar.addSeparator()
        self.addAction = QtGui.QAction(
            self.get_icon('list-add'), 'Add', self)
        self.addAction.setToolTip('Add a new argument')
        self.connect(self.addAction, QtCore.SIGNAL('triggered()'),
                     self.addArgument)
        self.toolBar.addAction(self.addAction)
        self.removeAction = QtGui.QAction(
            self.get_icon('list-remove'), 'Remove', self)
        self.removeAction.setToolTip('Remove the selected argument')
        self.connect(self.removeAction, QtCore.SIGNAL('triggered()'),
                     self.removeArgument)
        self.toolBar.addAction(self.removeAction)
        self.upAction = QtGui.QAction(
            self.get_icon('go-up'), 'Move up', self)
        self.upAction.setToolTip('Move argument up one position')
        self.connect(self.upAction, QtCore.SIGNAL('triggered()'),
                     self.moveUp)
        self.toolBar.addAction(self.upAction)
        self.downAction = QtGui.QAction(
            self.get_icon('go-down'), 'Move down', self)
        self.downAction.setToolTip('Move argument down one position')
        self.connect(self.downAction, QtCore.SIGNAL('triggered()'),
                     self.moveDown)
        self.toolBar.addAction(self.downAction)
        
        self.toolBar.addSeparator()

        self.showStdin = QtGui.QAction('stdin', self)
        self.showStdin.setToolTip('Check to use standard input as an input port')
        self.showStdin.setCheckable(True)
        self.toolBar.addAction(self.showStdin)
        self.showStdout = QtGui.QAction("stdout", self)
        self.showStdout.setToolTip('Check to use standard output as an output port')
        self.showStdout.setCheckable(True)
        self.toolBar.addAction(self.showStdout)
        self.showStderr = QtGui.QAction("stderr", self)
        self.showStderr.setToolTip('Check to use standard error as an output port')
        self.showStderr.setCheckable(True)
        self.toolBar.addAction(self.showStderr)
        self.envPort = QtGui.QAction("env", self)
        self.envPort.setToolTip('Check to add the "env" input port for specifying environment variables')
        self.envPort.setCheckable(True)
        self.toolBar.addAction(self.envPort)
        
        self.toolBar.addSeparator()

        self.stdAsFiles = QtGui.QAction('std file processing', self)
        self.stdAsFiles.setToolTip('Check to make pipes communicate using files instead of strings\nOnly useful when processing large files')
        self.stdAsFiles.setCheckable(True)
        self.toolBar.addAction(self.stdAsFiles)

        self.failWithCmd = QtGui.QAction('fail execution if return != 0', self)
        self.failWithCmd.setToolTip('If selected, VisTrails will check the exitcode, and abort the execution if not 0')
        self.failWithCmd.setCheckable(True)
        self.failWithCmd.setChecked(True)
        self.toolBar.addAction(self.failWithCmd)

        self.toolBar.addSeparator()

        self.previewPorts = QtGui.QAction('preview', self)
        self.previewPorts.setToolTip('Check which ports will be available for this module')
        self.connect(self.previewPorts, QtCore.SIGNAL('triggered()'),
                     self.preview_ports)
        self.toolBar.addAction(self.previewPorts)

        
        self.envOption = None
        
        self.commandLayout = QtGui.QHBoxLayout()
        self.commandLayout.setContentsMargins(5,5,5,5)
        tooltip = 'The command to execute'
        label = QtGui.QLabel("Command:")
        label.setFixedWidth(80)
        label.setToolTip(tooltip)
        self.commandLayout.addWidget(label)
        self.command = QtGui.QLineEdit()
        self.command.setToolTip(tooltip)
        self.commandLayout.addWidget(self.command)
        tooltip = 'Sets directory to execute from. Leave blank to ignore.'
        label = QtGui.QLabel("Directory:")
        label.setToolTip(tooltip)
        self.commandLayout.addWidget(label)
        self.dir = QtGui.QLineEdit()
        self.dir.setFixedWidth(140)
        self.dir.setToolTip(tooltip)
        self.commandLayout.addWidget(self.dir)
        self.vbox.addLayout(self.commandLayout)

        self.previewLayout = QtGui.QHBoxLayout()
        self.previewLayout.setContentsMargins(5,5,5,5)
        self.previewLayout.setAlignment(QtCore.Qt.AlignLeft)
        tooltip = 'Shows what the command will look like when executed in the command line'
        label = QtGui.QLabel("Preview:")
        label.setToolTip(tooltip)
        label.setFixedWidth(80)
        self.previewLayout.addWidget(label)
        self.preview = QtGui.QLabel()
        self.preview.setToolTip(tooltip)
        self.preview.setMaximumWidth(600)
        self.previewLayout.addWidget(self.preview)
        self.vbox.addLayout(self.previewLayout)

        self.importLayout = QtGui.QHBoxLayout()
        self.importLayout.setContentsMargins(5,5,5,5)
        self.importLayout.setAlignment(QtCore.Qt.AlignLeft)
        self.importLayout.addWidget(QtGui.QLabel("Man page:"))
        self.viewManButton = QtGui.QPushButton("view")
        self.viewManButton.setToolTip('View the man page for the current command')
        self.connect(self.viewManButton, QtCore.SIGNAL('clicked()'),
                     self.viewManPage)
        self.importLayout.addWidget(self.viewManButton)
        self.importManButton = QtGui.QPushButton("import")
        self.importManButton.setToolTip('Import arguments from the man page for the current command')
        self.connect(self.importManButton, QtCore.SIGNAL('clicked()'),
                     self.generateFromManPage)
        self.importLayout.addWidget(self.importManButton)

        self.importLayout.addWidget(QtGui.QLabel("help page (-h):"))
        self.viewHelpButton = QtGui.QPushButton("view")
        self.viewHelpButton.setToolTip('View the help (-h) page for the current command')
        self.connect(self.viewHelpButton, QtCore.SIGNAL('clicked()'),
                     self.viewHelpPage)
        self.importLayout.addWidget(self.viewHelpButton)
        self.importHelpButton = QtGui.QPushButton("import")
        self.importHelpButton.setToolTip('Import arguments from the help (-h) page for the current command')
        self.connect(self.importHelpButton, QtCore.SIGNAL('clicked()'),
                     self.generateFromHelpPage)
        self.importLayout.addWidget(self.importHelpButton)

        self.importLayout.addWidget(QtGui.QLabel("help page (--help):"))
        self.viewHelpButton2 = QtGui.QPushButton("view")
        self.viewHelpButton2.setToolTip('View the help (--help) page for the current command')
        self.connect(self.viewHelpButton2, QtCore.SIGNAL('clicked()'),
                     self.viewHelpPage2)
        self.importLayout.addWidget(self.viewHelpButton2)
        self.importHelpButton2 = QtGui.QPushButton("import")
        self.importHelpButton2.setToolTip('Import arguments from the help (--help) page for the current command')
        self.connect(self.importHelpButton2, QtCore.SIGNAL('clicked()'),
                     self.generateFromHelpPage2)
        self.importLayout.addWidget(self.importHelpButton2)
        self.vbox.addLayout(self.importLayout)

        self.stdinWidget = QArgWidget('stdin')
        self.stdinGroup = QtGui.QGroupBox('Standard input')
        self.stdinGroup.setLayout(QtGui.QHBoxLayout())
        self.stdinGroup.layout().addWidget(self.stdinWidget)
        self.layout().addWidget(self.stdinGroup)
        self.stdinGroup.setVisible(False)
        self.stdoutWidget = QArgWidget('stdout')
        self.stdoutGroup = QtGui.QGroupBox('Standard output')
        self.stdoutGroup.setLayout(QtGui.QHBoxLayout())
        self.stdoutGroup.layout().addWidget(self.stdoutWidget)
        self.layout().addWidget(self.stdoutGroup)
        self.stdoutGroup.setVisible(False)
        self.stderrWidget = QArgWidget('stderr')
        self.stderrGroup = QtGui.QGroupBox('Standard error')
        self.stderrGroup.setLayout(QtGui.QHBoxLayout())
        self.stderrGroup.layout().addWidget(self.stderrWidget)
        self.layout().addWidget(self.stderrGroup)
        self.stderrGroup.setVisible(False)

        self.connect(self.showStdin, QtCore.SIGNAL('toggled(bool)'),
                     self.stdinGroup.setVisible)
        self.connect(self.showStdout, QtCore.SIGNAL('toggled(bool)'),
                     self.stdoutGroup.setVisible)
        self.connect(self.showStderr, QtCore.SIGNAL('toggled(bool)'),
                     self.stderrGroup.setVisible)
        
        self.argList = QtGui.QListWidget()
        self.layout().addWidget(self.argList)

    def get_icon(self, name):
        return QtGui.QIcon(os.path.join(os.path.dirname(__file__),
                                        "icons",
                                        "%s.png" % name))
    def setTitle(self, file=None):
        self.parent().setWindowTitle("CLTools Wizard - " + (file if file else "untitled"))

    def newFile(self):
        self.conf = None
        self.file = None
        self.command.clear()
        self.dir.clear()
        self.showStdin.setChecked(False)
        self.showStdout.setChecked(False)
        self.showStderr.setChecked(False)
        self.envPort.setChecked(False)
        self.envOption = None
        while self.argList.count():
            item = self.argList.item(0)
            itemWidget = self.argList.itemWidget(item)
            itemWidget.setVisible(False)
            itemWidget.setParent(None)
            self.argList.removeItemWidget(item)
            item.setHidden(True)
            self.argList.takeItem(0)
        self.argList.hide()
        self.layout().takeAt(self.layout().indexOf(self.argList))
        self.argList = QtGui.QListWidget()
        self.layout().addWidget(self.argList)
        self.stdAsFiles.setChecked(False)
        self.failWithCmd.setChecked(True)
        self.setTitle()
        self.generate_preview()
    
    def openFile(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self,
                "Open Wrapper",
                self.file if self.file else default_dir(),
                "Wrappers (*%s)" % SUFFIX)
        if not fileName:
            return
        try:
            conf = json.load(open(fileName))
        except  ValueError as exc:
            print "Error opening Wrapper '%s': %s" % (fileName, exc)
            return
        self.newFile()
        self.file = fileName
        self.setTitle(self.file)
        self.command.setText(conf.get('command', ''))
        self.dir.setText(conf.get('dir', ''))
        if 'stdin' in conf:
            self.stdinWidget.fromList(conf['stdin'])
        self.stdinGroup.setVisible('stdin' in conf)
        self.showStdin.setChecked('stdin' in conf)
        if 'stdout' in conf:
            self.stdoutWidget.fromList(conf['stdout'])
        self.stdoutGroup.setVisible('stdout' in conf)
        self.showStdout.setChecked('stdout' in conf)
        if 'stderr' in conf:
            self.stderrWidget.fromList(conf['stderr'])
        self.stderrGroup.setVisible('stderr' in conf)
        self.showStderr.setChecked('stderr' in conf)
        for argConf in conf.get('args', []):
            arg = QArgWidget()
            arg.fromList(argConf)
            item = QtGui.QListWidgetItem()
            item.setSizeHint(arg.sizeHint())
            self.argList.addItem(item)
            self.argList.setItemWidget(item, arg)
        self.envPort.setChecked('options' in conf and
                                'env_port' in conf['options'])
        self.stdAsFiles.setChecked('options' in conf and
                                   'std_using_files' in conf['options'])
        self.failWithCmd.setChecked('options' in conf and
                                    'fail_with_cmd' in conf['options'])
        self.envOption = conf['options']['env'] \
                 if ('options' in conf and 'env' in conf['options']) else None
        self.conf = conf
        self.generate_preview()
            
    def get_current_conf(self):
        conf = {}
        conf['command'] = self.command.text()
        dir = self.dir.text()
        if dir:
            conf['dir'] = dir
            
        if self.stdinGroup.isVisible():
            conf['stdin'] = self.stdinWidget.toList()
        if self.stdoutGroup.isVisible():
            conf['stdout'] = self.stdoutWidget.toList()
        if self.stderrGroup.isVisible():
            conf['stderr'] = self.stderrWidget.toList()
        args = []
        for row in xrange(self.argList.count()):
            arg = self.argList.itemWidget(self.argList.item(row))
            args.append(arg.toList())
        conf['args'] = args

        options = {}
        if self.stdAsFiles.isChecked():
            options['std_using_files'] = ''
        if self.failWithCmd.isChecked():
            options['fail_with_cmd'] = ''
        if self.envPort.isChecked():
            options['env_port'] = ''
        if self.envOption:
            options['env'] = self.envOption
        if options:
            conf['options'] = options
        return conf
    
    def save(self):
        if not self.file:
            self.saveAs()
            if not self.file:
                return
        self.conf = self.get_current_conf()
        f = open(self.file, "w")
        json.dump(self.conf, f, sort_keys=True, indent=4)
        f.close()
        self.generate_preview()

    def saveAs(self):
        fileName = QtGui.QFileDialog.getSaveFileName(self,
                            "Save Wrapper as",
                            self.file if self.file else default_dir(),
                            "Wrappers (*%s)" % SUFFIX)
        if fileName:
            self.file = fileName
            if not self.file.endswith(SUFFIX):
                self.file += SUFFIX
            self.save()
            self.setTitle(self.file)

    def refresh(self):
        self.save()
        self.reload_scripts()
    
    def generate_preview(self):
        # generate preview from self.conf
        c = self.get_current_conf()
        if not c:
            self.preview.setText('')
            self.preview.setToolTip('')
            return
        
        text = ''

        # show env as bash-style prefix
        if 'options' in c:
            o = c['options']
            # env_port
            if 'env_port' in o:
                text += '<env_port>'
            # env
            if 'env' in o:
                if text:
                    text += ' '
                text += c['options']['env']
            if text:
                text += ' '

        # command
        text += c['command']

        # args
        if 'args' in c:
            for type, name, klass, opts in c['args']:
                type = type.lower()
                klass = klass.lower()
                text += ' '
                if type == 'constant':
                    if 'flag' in opts:
                        text += opts['flag'] + ' '
                    text += quote_arg(name)
                    continue
                if 'required' not in opts:
                    text += '['
                if klass in 'list':
                    text += '{'
                if 'flag' in opts:
                    text += opts['flag']
                    if klass != 'flag':
                        text += ' '
                if 'prefix' in opts:
                    text += opts['prefix']
                if type!='input' or klass != 'flag':
                    text += '<'
                if klass == 'list':
                    text += opts['type'].lower() \
                            if ('type' in opts and opts['type']) else 'string'
                elif type=='input' and klass == 'flag':
                    if 'flag' not in opts:
                        text += quote_arg(name)
                elif type in ['output', 'inputoutput']:
                    text += 'file'
                else:
                    text += klass
                if type!='input' or klass != 'flag':
                    text += '>'
                if klass == 'list':
                    text += '}'
                if 'required' not in opts:
                    text += ']' 
                
        self.preview.setText(text)
        self.preview.setToolTip(text)

    def preview_ports(self):
        # show dialog with the ports that this module will have
        self.generate_preview()
        conf = self.get_current_conf()
        if not conf:
            return

        intext = []
        outtext = []

        if 'stdin' in conf:
            name, type, options = conf['stdin']
            optional = " (visible)" if "required" in options else ""
            intext.append("%s: %s%s" % (name, type, optional))
        if 'stdout' in conf:
            name, type, options = conf['stdout']
            optional = " (visible)" if "required" in options else ""
            outtext.append("%s: %s%s" % (name, type, optional))
        if 'stderr' in conf:
            name, type, options = conf['stderr']
            optional = " (visible)" if "required" in options else ""
            outtext.append("%s: %s%s" % (name, type, optional))
        if 'options' in conf and 'env_port' in conf['options']:
            intext.append('env: String')
        for type, name, klass, options in conf['args']:
            optional = " (visible)" if "required" in options else ""
            if 'input' == type.lower():
                intext.append("%s: %s%s" % (name, klass, optional))
            elif 'output' == type.lower():
                outtext.append("%s: %s%s" % (name, klass, optional))
            elif 'inputoutput' == type.lower():
                intext.append("%s: %s%s" % (name, 'File', optional))
                outtext.append("%s: %s%s" % (name, 'File', optional))
        
        intext = ''.join(['Input %s. %s\n' % (i+1, t)
                          for i, t in zip(xrange(len(intext)), intext)])
        outtext = ''.join(['Output %s. %s\n' % (i+1, t)
                            for i, t in zip(xrange(len(outtext)), outtext)])
        
        self.helppageView = QManpageDialog("Module Ports for this Wrapper",
                                     intext + "\n" + outtext, self)
        self.helppageView.show()
    
    def loadFromCommand(self, argv):
        self.command.setText(argv[0])
        pos = 0
        for argName in argv[1:]:
            arg = QArgWidget()
            arg.guess(argName, pos)
            pos += 1
            item = QtGui.QListWidgetItem()
            item.setSizeHint(arg.sizeHint())
            self.argList.addItem(item)
            self.argList.setItemWidget(item, arg)

    def addArgument(self):
        arg = QArgWidget()
        item = QtGui.QListWidgetItem()
        item.setSizeHint(arg.sizeHint())
        self.argList.addItem(item)
        self.argList.setItemWidget(item, arg)

    def moveUp(self):
        currentRow = self.argList.currentRow()
        if currentRow < 1:
            return
        # moving widgets does not work so we just switch contents
        w0 = self.argList.itemWidget(self.argList.item(currentRow-1))
        w1 = self.argList.itemWidget(self.argList.item(currentRow))
        arg0 = w0.toList()
        arg1 = w1.toList()
        w0.fromList(arg1)
        w1.fromList(arg0)
        self.argList.setCurrentRow(currentRow - 1)

    def moveDown(self):
        currentRow = self.argList.currentRow()
        if currentRow<0 or currentRow>self.argList.count()-2:
            return
        # moving widgets does not work so we just switch contents
        w1 = self.argList.itemWidget(self.argList.item(currentRow))
        w2 = self.argList.itemWidget(self.argList.item(currentRow+1))
        arg1 = w1.toList()
        arg2 = w2.toList()
        w1.fromList(arg2)
        w2.fromList(arg1)
        self.argList.setCurrentRow(currentRow + 1)

    def removeArgument(self):
        currentRow = self.argList.currentRow()
        if currentRow<0:
            return
        self.argList.takeItem(currentRow)

    def keyPressEvent(self, event):
        if event.key() in [QtCore.Qt.Key_Delete, QtCore.Qt.Key_Backspace]:
            self.removeArgument()
        else:
            QtGui.QWidget.keyPressEvent(self, event)
    
    def parse(self, text):
        """ parse(self, text)
        parses the description of a command and extracts possible arguments
        works on both man pages and help (-h) pages
        """
        lines = text.split('\n')
        args = []
        arg = None
        for line in lines:
            line = line.strip()
            if len(line)>1 and line[0] == '-':
                # make sure description is removed
                flag = line.split('   ')[0].strip()
                # use the last in the list of equal flags
                #it is assumed to be most descriptive
                flag = flag.split(',')[-1].strip()
                # remove any special characters after flag
                for i in ['=', '[', ' ']:
                    flag = flag.split(i)[0].strip()
                # new arg (type, flag, description)
                arg = ['Flag', flag, line]
                if '=' in line: # assume string attribute with prefix 
                    arg[0] = 'String'
                args.append(arg)
            elif arg:
                if not line:
                    arg = None
                else:
                    arg[2] += ' %s' % line.strip()
        #print "args:"
        #for arg in args:
        #    print arg
        return args
        
    def runProcess(self, args):
        try:
            command = Command(args)
            status, text, stderr = command.run(stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               shell=True)
            if not (text and len(text)):
                text = stderr
                if not (text and len(text)) or (text and text.startswith('No ')):
                    return None
            # fix weird formatting
            for a, b in encode_list:
                text = text.replace(a, b)
            return text
        except Exception:
            return None
    
    def generateFromManPage(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', 'man %s | col -b' % command])
        if not text:
            QtGui.QMessageBox.warning(self, "Man page not found",
                                      "For command '%s'" % command)
            return
        args = self.parse(text)
        title = "Import arguments from man page for '%s'" % command
        self.manpageImport = QManpageImport(title, args, self)
        self.connect(self.manpageImport,
                       QtCore.SIGNAL("importArgs(PyQt_PyObject)"),
                       self.importArgs)
        self.manpageImport.show()

    def generateFromHelpPage(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', command + ' -h'])
        if not text:
            QtGui.QMessageBox.warning(self, "Help page (-h) not found",
                                      "For command '%s'" % command)
            return
        args = self.parse(text)

        title = "Import arguments from help page (-h) for '%s'" % command
        self.helppageImport = QManpageImport(title, args, self)
        self.connect(self.helppageImport,
                       QtCore.SIGNAL("importArgs(PyQt_PyObject)"),
                       self.importArgs)
        self.helppageImport.show()

    def generateFromHelpPage2(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', command + ' --help'])
        if not text:
            QtGui.QMessageBox.warning(self, "Help page (--help) not found",
                                      "For command '%s'" % command)
            return
        args = self.parse(text)

        title = "Import arguments from help page (--help) for '%s'" % command
        self.helppageImport = QManpageImport(title, args, self)
        self.connect(self.helppageImport,
                       QtCore.SIGNAL("importArgs(PyQt_PyObject)"),
                       self.importArgs)
        self.helppageImport.show()

    def viewManPage(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', 'man %s | col -b' % command])
        if not text:
            QtGui.QMessageBox.warning(self, "Man page not found",
                                      "For command '%s'" % command)
            return
        title = "man page for '%s'" % command
        self.manpageView = QManpageDialog(title, text, self)
        self.manpageView.show()

    def viewHelpPage(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', command + ' -h'])
        if not text:
            QtGui.QMessageBox.warning(self, "Help page (-h) not found",
                                      "For command '%s'" % command)
            return
        title = "Help page for '%s'" % command
        self.helppageView = QManpageDialog(title, text, self)
        self.helppageView.show()

    def viewHelpPage2(self):
        command = self.command.text()
        if command == '':
            return
        text = self.runProcess(['-c', command + ' --help'])
        if not text:
            QtGui.QMessageBox.warning(self, "Help page (--help) not found",
                                      "For command '%s'" % command)
            return
        title = "Help page for '%s'" % command
        self.helppageView = QManpageDialog(title, text, self)
        self.helppageView.show()

    def importArgs(self, args):
        for arg in args:
            item = QtGui.QListWidgetItem()
            item.setSizeHint(arg.sizeHint())
            self.argList.insertItem(0, item)
            self.argList.setItemWidget(item, arg)

class QArgWidget(QtGui.QWidget):
    """ Widget for configuring an argument """
    KLASSES = {
            'input': ['flag', 'file', 'path', 'directory',
                      'string', 'integer', 'float', 'list'],
            'output': ['file', 'string'],
            'inputoutput': ['file'],
            'stdin': ['file', 'string'],
            'stdout': ['file', 'string'],
            'stderr': ['file', 'string'],
        }
    KLASSNAMES = {
            'flag': 'Boolean flag',
            'string': 'String',
            'integer': 'Integer',
            'float': 'Float',
            'file': 'File',
            'path': 'Path',
            'directory': 'Directory',
            'list': 'List',
        }

    TYPES = ['input', 'output', 'inputoutput', 'constant']
    TYPENAMES = {
            'input': 'Input Port',
            'output': 'Output Port',
            'inputoutput': 'InputOutput Port',
            'constant': 'Constant',
        }

    def __init__(self, argtype='Input', name='untitled', klass='Flag', options={}, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.stdTypes = ['stdin', 'stdout', 'stderr']
        self.stdLabels = ['Standard input', 'Standard output', 'Standard error']
        self.stdDict = dict(zip(self.stdTypes, self.stdLabels))

        self.argtype = argtype.lower()
        self.name = name
        self.klass = klass.lower()
        self.options = options

        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        self.buildWidget()

    def buildWidget(self):
        layout = self.layout()
        layout.setContentsMargins(2,2,2,2)
        # remove any previous layout
        layout1 = QtGui.QHBoxLayout()
        layout1.setContentsMargins(2,2,2,2)
        layout.addLayout(layout1)
        if self.argtype not in self.stdTypes:
            layout2 = QtGui.QHBoxLayout()
            layout2.setContentsMargins(2,2,2,2)
            layout.addLayout(layout2)
        else:
            layout2 = layout1
        # type of argument
        if self.argtype not in self.stdTypes:
            self.typeList = QtGui.QComboBox()
            self.typeDict = {}
            for i, n in enumerate(self.TYPES):
                self.typeList.addItem(self.TYPENAMES[n], n)
                self.typeDict[n] = i
            self.typeList.setCurrentIndex(self.typeDict.get(self.argtype, 0))
            #label = QtGui.QLabel('Type:')
            tt = "Select if argument will be an input port, output port, both, or a hidden constant. InputOutput's are always files."
            #label.setToolTip(tt)
            self.typeList.setToolTip(tt)
            #layout1.addWidget(label)
            layout1.addWidget(self.typeList)
        else:
            self.typeList = None
        # type of port
        self.klassList = QtGui.QComboBox()
        klasses = self.KLASSES[self.argtype]
        self.klassDict = {}
        for i, n in enumerate(klasses):
            self.klassList.addItem(self.KLASSNAMES[n], n)
            self.klassDict[n] = i
        self.klassList.setCurrentIndex(self.klassDict.get(self.klass, 0))
        #label = QtGui.QLabel('Class:')
        tt = 'Port Type. Can be String, Integer, Float, File/Directory/Path or Boolean Flag. List means an input list of one of the other types. Only File and String should be used for output ports.'
        self.klassList.setToolTip(tt)
        #label.setToolTip(tt)
        #layout1.addWidget(label)
        layout1.addWidget(self.klassList)
        # name of port
        self.nameLine = QtGui.QLineEdit(self.name)
        label = QtGui.QLabel('Name:')
        tt = 'Name of the port, or the value for constants'
        label.setToolTip(tt)
        self.nameLine.setToolTip(tt)
        layout1.addWidget(label)
        layout1.addWidget(self.nameLine)
        # options are different for each widget
        if self.argtype not in self.stdTypes:
            # all args can have flag
            self.flag = QtGui.QLineEdit(self.options.get('flag', ''))
            label = QtGui.QLabel('Flag:')
            tt = 'a short-style flag before your input. Example: "-f" -> "-f yourinput"'
            label.setToolTip(tt)
            self.flag.setToolTip(tt)
            self.flag.setFixedWidth(100)
            layout1.addWidget(label)
            layout1.addWidget(self.flag)
        
            # all args can have prefix
            self.prefix = QtGui.QLineEdit(self.options.get('prefix', ''))
            label = QtGui.QLabel('Prefix:')
            tt = 'a long-style prefix to your input. Example: "--X=" -> "--X=yourinput"'
            label.setToolTip(tt)
            self.prefix.setToolTip(tt)
            layout1.addWidget(label)
            layout1.addWidget(self.prefix)

        # all can be required
        self.required = QtGui.QCheckBox()
        self.required.setChecked('required' in self.options)
        label = QtGui.QLabel('Visible:')
        tt = 'Check to make port always visible in VisTrails'
        label.setToolTip(tt)
        self.required.setToolTip(tt)
        layout2.addWidget(label)
        layout2.addWidget(self.required)
        
        # subtype
        self.subList = ['String', 'Integer', 'Float', 'File', 'Directory', 'Path']
        self.subDict = dict(zip(self.subList, xrange(len(self.subList))))
        self.subDict.update(dict(zip([s.lower() for s in self.subList], xrange(len(self.subList)))))
        self.subtype = QtGui.QComboBox()
        self.subtype.addItems(self.subList)
        self.subtype.setCurrentIndex(self.subDict.get(self.options.get('type', 'String'), 0))
        self.listLabel = QtGui.QLabel('List type:')
        self.subtype.setVisible(False)
        tt = 'Choose type of values in List'
        self.subtype.setToolTip(tt)
        self.listLabel.setToolTip(tt)
        layout2.addWidget(self.listLabel)
        layout2.addWidget(self.subtype)
        self.listLabel.setVisible(False)
        self.subtype.setVisible(False)
        
        # input files and inputoutput's can set file suffix
        self.suffix = QtGui.QLineEdit(self.options.get('suffix', ''))
        self.suffixLabel = QtGui.QLabel('File suffix:')
        tt = 'Sets the specified file ending on the created file, like for example: ".txt"'
        self.suffixLabel.setToolTip(tt)
        self.suffix.setToolTip(tt)
        self.suffix.setFixedWidth(50)
        layout2.addWidget(self.suffixLabel)
        layout2.addWidget(self.suffix)

        self.typeChanged()
        self.klassChanged()

        # description
        self.desc = QtGui.QLineEdit(self.options.get('desc', ''))
        label = QtGui.QLabel('Description:')
        tt = 'Add a helpful description of the port'
        label.setToolTip(tt)
        self.desc.setToolTip(tt)
        layout2.addWidget(label)
        layout2.addWidget(self.desc)
        
        if self.argtype not in self.stdTypes:
            self.connect(self.klassList, QtCore.SIGNAL('currentIndexChanged(int)'),
                     self.klassChanged)
            self.connect(self.typeList, QtCore.SIGNAL('currentIndexChanged(int)'),
                     self.typeChanged)

    def getValues(self):
        """ get the values from the widgets and store them """
        self.klass = self.klassList.itemData(self.klassList.currentIndex())
        self.name = self.nameLine.text()
        self.options = {}
        if self.argtype not in self.stdTypes:
            flag = self.flag.text()
            if flag:
                self.options['flag'] = flag
            prefix = self.prefix.text()
            if prefix:
                self.options['prefix'] = prefix
        desc = self.desc.text()
        if desc:
            self.options['desc'] = desc
        if self.required.isChecked():
            self.options['required'] = ''
        if self.klass == 'list':
            subtype = self.subtype.currentText()
            if subtype:
                self.options['type'] = subtype
        type = self.argtype.lower()
        suffix = self.suffix.text()
        if (type == 'output' or type == 'inputoutput') and suffix:
            self.options['suffix'] = suffix

    def setValues(self):
        if self.argtype not in self.stdTypes:
            self.typeList.setCurrentIndex(self.typeDict.get(self.argtype, 0))
        self.nameLine.setText(self.name)
        self.klassList.setCurrentIndex(self.klassDict[self.klass])
        if self.argtype not in self.stdTypes:
            self.flag.setText(self.options.get('flag', ''))
            self.prefix.setText(self.options.get('prefix', ''))
            self.subtype.setCurrentIndex(self.subDict.get(self.options.get('type', 'String')))
            self.subtype.setCurrentIndex(self.subDict.get(self.options.get('type', 'String'), 0))
        self.required.setChecked('required' in self.options)
        self.desc.setText(self.options.get('desc', ''))
        type = self.argtype.lower()
        if type == 'output' or type == 'inputoutput':
            self.suffix.setText(self.options.get('suffix', ''))
        self.typeChanged()
        self.klassChanged()

    def toList(self):
        self.getValues()
        if self.argtype not in self.stdTypes:
            return [self.argtype, self.name, self.klass, dict(self.options)]
        else:
            return [self.name, self.klass, dict(self.options)]

    def fromList(self, arg):
        if self.argtype not in self.stdTypes:
            self.argtype, self.name, klass, self.options = arg
        else:
            self.name, klass, self.options = arg
        self.klass = klass.lower()
        self.setValues()

    def klassChanged(self, index=None):
        if self.argtype in self.stdTypes:
            return
        klass = self.klassList.itemData(self.klassList.currentIndex())
        type = self.typeList.itemData(self.typeList.currentIndex())
        self.listLabel.setVisible(klass == "list" and type == 'input')
        self.subtype.setVisible(klass == "list" and type == 'input')

    def typeChanged(self, index=None):
        if self.argtype in self.stdTypes:
            return
        type = self.typeList.itemData(self.typeList.currentIndex())
        if index is not None and type == self.argtype:
            return
        self.argtype = type
        if type in ('constant', 'inputoutput'):
            self.klassList.hide()
        else:
            self.klassList.show()
            self.klassList.clear()
            klasses = self.KLASSES[self.argtype]
            self.klassDict = {}
            for i, n in enumerate(klasses):
                self.klassList.addItem(self.KLASSNAMES[n], n)
                self.klassDict[n] = i
            self.klassList.setCurrentIndex(self.klassDict.get(self.klass, 0))
        self.suffixLabel.setVisible(type == 'output' or type == 'inputoutput')
        self.suffix.setVisible(type == 'output' or type == 'inputoutput')

    def guess(self, name, count=0):
        """ add argument by guessing what the arg might be """
        if '.' in name or '/' in name or '\\' in name: # guess path
            if os.path.isfile(name):
                guessed, type_ = 'file', 'File'
            elif os.path.isdir(name):
                guessed, type_ = 'directory', 'Directory'
            else:
                guessed, type_ = 'path', 'Path'
            self.fromList(['Input', '%s%d' % (guessed, count), type_,
                           {'desc':'"%s" guessed to be an Input %s' % (name, guessed)}])
        elif name.startswith('-'): # guess flag
            self.fromList(['Input', 'flag%s' % name, 'Flag',
                           {'desc':'"%s" guessed to be a flag' % name,
                            'flag':name}])
        else: # guess string
            self.fromList(['Input', 'input%s' % count, 'String',
                           {'desc':'"%s" guessed to be an input string' % name}])
            

class QManpageDialog(QtGui.QDialog):
    def __init__(self, title, text, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        text = "<pre>%s</pre>" % text
        self.textEdit = QtGui.QTextEdit(text)
        self.textEdit.setReadOnly(True)
        self.setLayout(QtGui.QHBoxLayout())
        self.layout().addWidget(self.textEdit)
        self.resize(800,600)

class QManpageImport(QtGui.QDialog):
    def __init__(self, title, args, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        self.argLayout = QtGui.QVBoxLayout()
        for arg in args:
            if arg[0] == 'Flag':
                w = QArgWidget(name=arg[1], options={'flag':arg[1],
                                                     'desc':arg[2]})
            else:
            #if arg[0] == 'String':
                w = QArgWidget(klass='String', name=arg[1],
                               options={'prefix':arg[1] + '=',
                                        'desc':arg[2]})
            widgetLayout = QtGui.QHBoxLayout()
            widgetLayout.addWidget(QtGui.QCheckBox())
            widgetLayout.addWidget(w)
            self.argLayout.addLayout(widgetLayout)
        scroll = QtGui.QScrollArea()
        w = QtGui.QWidget()
        w.setLayout(self.argLayout)
        scroll.setWidget(w)
        scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        scroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        scroll.setWidgetResizable(True)
        self.setLayout(QtGui.QVBoxLayout())
        self.layout().addWidget(scroll)
        layout2 = QtGui.QHBoxLayout()
        self.layout().addLayout(layout2)
        self.closeButton = QtGui.QPushButton('Close')
        self.closeButton.setToolTip('Close this window')
        self.connect(self.closeButton, QtCore.SIGNAL('clicked()'),
                     self.close)
        layout2.addWidget(self.closeButton)
        self.selectAllButton = QtGui.QPushButton('Select All')
        self.selectAllButton.setToolTip('Select All arguments')
        self.connect(self.selectAllButton, QtCore.SIGNAL('clicked()'),
                     self.selectAll)
        layout2.addWidget(self.selectAllButton)
        self.selectNoneButton = QtGui.QPushButton('Select None')
        self.selectNoneButton.setToolTip('Unselect All arguments')
        self.connect(self.selectNoneButton, QtCore.SIGNAL('clicked()'),
                     self.selectNone)
        layout2.addWidget(self.selectNoneButton)
        self.addSelectedButton = QtGui.QPushButton('Import Selected')
        self.addSelectedButton.setToolTip('Import all selected arguments')
        self.connect(self.addSelectedButton, QtCore.SIGNAL('clicked()'),
                     self.addSelected)
        layout2.addWidget(self.addSelectedButton)
        self.resize(800,600)

    def selectAll(self):
        for i in xrange(self.argLayout.count()):
            w = self.argLayout.layout().itemAt(i)
            w.layout().itemAt(0).widget().setChecked(True)

    def selectNone(self):
        for i in xrange(self.argLayout.count()):
            w = self.argLayout.layout().itemAt(i)
            w.layout().itemAt(0).widget().setChecked(False)

    def addSelected(self):
        # collect selected arguments and send through signal
        args = []
        remove_list = []
        for i in xrange(self.argLayout.count()):
            w = self.argLayout.layout().itemAt(i)
            if w.layout().itemAt(0).widget().isChecked():
                remove_list.append(w)
                args.append(w.layout().itemAt(1).widget())
        for w in remove_list:
            w.layout().itemAt(0).widget().setChecked(False)
            w.layout().itemAt(0).widget().hide()
            w.layout().itemAt(1).widget().hide()
            self.argLayout.removeItem(w)

        self.emit(QtCore.SIGNAL('importArgs(PyQt_PyObject)'), args)

class QCLToolsWizardWindow(QtGui.QMainWindow):

    def __init__(self, parent=None, reload_scripts=None):
        QtGui.QMainWindow.__init__(self, parent)
        self.wizard = QCLToolsWizard(self, reload_scripts)
        self.setCentralWidget(self.wizard)
        self.setWindowTitle("CLTools Wizard")
        self.resize(1000,600)
        
if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    window = QCLToolsWizardWindow()
    if len(sys.argv)>2 and sys.argv[1] == '-c':
        # read command from command line
        window.wizard.loadFromCommand(sys.argv[2:])
    window.show()
    app.exec_()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" The file describes the parameter exploration tab for VisTrails

QParameterExplorationTab
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from ast import literal_eval
from xml.dom.minidom import parseString
from xml.sax.saxutils import escape
from vistrails.core import debug
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.param_explore import ActionBasedParameterExploration
from vistrails.core.system import current_time, strftime, \
    get_vistrails_default_pkg_prefix
from vistrails.gui.common_widgets import QDockContainer, QToolWindowInterface
from vistrails.gui.paramexplore.pe_table import QParameterExplorationWidget, QParameterSetEditor
from vistrails.gui.paramexplore.virtual_cell import QVirtualCellWindow
from vistrails.gui.paramexplore.param_view import QParameterView
from vistrails.gui.paramexplore.pe_pipeline import QAnnotatedPipelineView

################################################################################

class QParameterExplorationTab(QDockContainer, QToolWindowInterface):
    """
    QParameterExplorationTab is a tab containing different widgets
    related to parameter exploration
    
    """
    explorationId = 0
    
    def __init__(self, parent=None):
        """ QParameterExplorationTab(parent: QWidget)
                                    -> QParameterExplorationTab
        Make it a main window with dockable area and a
        QParameterExplorationTable
        
        """
        QDockContainer.__init__(self, parent)
        self.setWindowTitle('Parameter Exploration')
        self.toolWindow().setFeatures(QtGui.QDockWidget.NoDockWidgetFeatures)
        self.toolWindow().hide()

        self.peWidget = QParameterExplorationWidget()
        self.setCentralWidget(self.peWidget)
        self.connect(self.peWidget.table,
                     QtCore.SIGNAL('exploreChange(bool)'),
                     self.exploreChange)

        self.paramView = QParameterView(self)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea,
                           self.paramView.toolWindow())
        
        self.annotatedPipelineView = QAnnotatedPipelineView(self)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea,
                           self.annotatedPipelineView.toolWindow())
        
        self.virtualCell = QVirtualCellWindow(self)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea,
                           self.virtualCell.toolWindow())
        
        self.controller = None
        self.currentVersion = -1

    def addViewActionsToMenu(self, menu):
        """addViewActionsToMenu(menu: QMenu) -> None
        Add toggle view actions to menu
        
        """
        menu.addAction(self.paramView.toolWindow().toggleViewAction())
        menu.addAction(self.annotatedPipelineView.toolWindow().toggleViewAction())
        menu.addAction(self.virtualCell.toolWindow().toggleViewAction())

    def removeViewActionsFromMenu(self, menu):
        """removeViewActionsFromMenu(menu: QMenu) -> None
        Remove toggle view actions from menu
        
        """
        menu.removeAction(self.paramView.toolWindow().toggleViewAction())
        menu.removeAction(self.annotatedPipelineView.toolWindow().toggleViewAction())
        menu.removeAction(self.virtualCell.toolWindow().toggleViewAction())
        
    def setController(self, controller):
        """ setController(controller: VistrailController) -> None
        Assign a controller to the parameter exploration tab
        
        """
        self.controller = controller

    def getParameterExploration(self):
        """ getParameterExploration() -> string
        Generates an XML string that represents the current
        parameter exploration, and which can be loaded with
        setParameterExploration().
        
        """
        # Construct xml for persisting parameter exploration
        escape_dict = { "'":"&apos;", '"':'&quot;', '\n':'&#xa;' }
        timestamp = strftime(current_time(), '%Y-%m-%d %H:%M:%S')
        # TODO: For now, we use the timestamp as the 'name' - Later, we should set 'name' based on a UI input field
        xml = '\t<paramexp dims="%s" layout="%s" date="%s" name="%s">' % (str(self.peWidget.table.label.getCounts()), str(self.virtualCell.getConfiguration()[2]), timestamp, timestamp)
        for i in xrange(self.peWidget.table.layout().count()):
            pEditor = self.peWidget.table.layout().itemAt(i).widget()
            if pEditor and isinstance(pEditor, QParameterSetEditor):
                firstParam = True
                for paramWidget in pEditor.paramWidgets:
                    paramInfo = paramWidget.param
                    interpolator = paramWidget.editor.stackedEditors.currentWidget()
                    intType = interpolator.exploration_name
                    # Write function tag prior to the first parameter of the function
                    if firstParam:
                        xml += '\n\t\t<function id="%s" alias="%s" name="%s">' % (paramInfo.parent_id, paramInfo.is_alias, pEditor.info[0])
                        firstParam = False
                    # Write parameter tag
                    xml += '\n\t\t\t<param id="%s" dim="%s" interp="%s"' % (paramInfo.id, paramWidget.getDimension(), intType)
                    if intType == 'Linear Interpolation':
                        xml += ' min="%s" max="%s"' % (interpolator.fromEdit.get_value(), interpolator.toEdit.get_value())
                    elif intType == 'List':
                        xml += ' values="%s"' % escape(str(interpolator._str_values), escape_dict)
                    elif intType == 'User-defined Function':
                        xml += ' code="%s"' % escape(interpolator.function, escape_dict)
                    xml += '/>'
                xml += '\n\t\t</function>'
        xml += '\n\t</paramexp>'
        return xml

    def setParameterExploration(self, xmlString):
        """ setParameterExploration(xmlString: string) -> None
        Sets the current parameter exploration to the one
        defined by 'xmlString'.
        
        """
        if not xmlString:
            return
        # Parse/validate the xml
        try:
            xmlDoc = parseString(xmlString).documentElement
        except Exception:
            debug.critical("Parameter Exploration load failed because of "
                           "invalid XML:\n\n%s" % xmlString)
            return
        # Set the exploration dimensions
        dims = literal_eval(xmlDoc.attributes['dims'].value)
        self.peWidget.table.label.setCounts(dims)
        # Set the virtual cell layout
        layout = literal_eval(xmlDoc.attributes['layout'].value)
        self.virtualCell.setConfiguration(layout)
        # Populate parameter exploration window with stored functions and aliases
        for f in xmlDoc.getElementsByTagName('function'):
            # Retrieve function attributes
            f_id = long(f.attributes['id'].value)
            f_is_alias = (str(f.attributes['alias'].value) == 'True')
            # Search the parameter treeWidget for this function and add it directly
            newEditor = None
            for tidx in xrange(self.paramView.treeWidget.topLevelItemCount()):
                moduleItem = self.paramView.treeWidget.topLevelItem(tidx)
                for cidx in xrange(moduleItem.childCount()):
                    paramInfo = moduleItem.child(cidx).parameter
                    name, params = paramInfo
                    if params[0].parent_id == f_id and params[0].is_alias == f_is_alias:
                        newEditor = self.peWidget.table.addParameter(paramInfo)
            # Retrieve params for this function and set their values in the UI
            if newEditor:
                for p in f.getElementsByTagName('param'):
                    # Locate the param in the newly added param editor and set values
                    p_id = long(p.attributes['id'].value)
                    for paramWidget in newEditor.paramWidgets:
                        if paramWidget.param.id == p_id:
                            # Set Parameter Dimension (radio button)
                            p_dim = int(p.attributes['dim'].value)
                            paramWidget.setDimension(p_dim)
                            # Set Interpolator Type (dropdown list)
                            p_intType = str(p.attributes['interp'].value)
                            paramWidget.editor.selectInterpolator(p_intType)
                            # Set Interpolator Value(s)
                            interpolator = paramWidget.editor.stackedEditors.currentWidget()
                            if p_intType == 'Linear Interpolation':
                                # Set min/max
                                p_min = str(p.attributes['min'].value)
                                p_max = str(p.attributes['max'].value)
                                interpolator.fromEdit.setText(p_min)
                                interpolator.toEdit.setText(p_max)
                            elif p_intType == 'List':
                                p_values = str(p.attributes['values'].value)
                                # Set internal list structure
                                interpolator._str_values = \
                                        literal_eval(p_values)
                                # Update UI list
                                if interpolator.type == 'String':
                                    interpolator.listValues.setText(p_values)
                                else:
                                    interpolator.listValues.setText(p_values.replace("'", "").replace('"', ''))
                            elif p_intType == 'User-defined Function':
                                # Set function code
                                p_code = str(p.attributes['code'].value)
                                interpolator.function = p_code

    def showEvent(self, event):
        """ showEvent(event: QShowEvent) -> None
        Update the tab when it is shown
        
        """
        if self.currentVersion!=self.controller.current_version:
            self.currentVersion = self.controller.current_version
            # Update the virtual cell
            pipeline = self.controller.current_pipeline
            self.virtualCell.updateVirtualCell(pipeline)

            # Now we need to inspect the parameter list
            self.paramView.treeWidget.updateFromPipeline(pipeline)

            # Update the annotated ids
            self.annotatedPipelineView.updateAnnotatedIds(pipeline)

            # Update the parameter exploration table
            self.peWidget.updatePipeline(pipeline)

            # Update the UI with the most recent parameter exploration
            # TODO: For now, we just strip the root tags since there's only one
            #       exploration - Later we should parse the root tree and select
            #       the active exploration based on date, or user choice
            xmlString = self.controller.vistrail.get_paramexp(self.currentVersion)
            if xmlString is not None:
                striplen = len("<paramexps>")
                xmlString = xmlString[striplen:-(striplen+1)].strip()
                self.setParameterExploration(xmlString)

    def performParameterExploration(self):
        """ performParameterExploration() -> None        
        Perform the exploration by collecting a list of actions
        corresponding to each dimension
        
        """
        registry = get_module_registry()
        actions = self.peWidget.table.collectParameterActions()
        spreadsheet_pkg = 'org.vistrails.vistrails.spreadsheet'
        # Set the annotation to persist the parameter exploration
        # TODO: For now, we just replace the existing exploration - Later we should append them.
        xmlString = "<paramexps>\n" + self.getParameterExploration() + "\n</paramexps>"
        self.controller.vistrail.set_paramexp(self.currentVersion, xmlString)
        self.controller.set_changed(True)

        if self.controller.current_pipeline and actions:
            explorer = ActionBasedParameterExploration()
            (pipelines, performedActions) = explorer.explore(
                self.controller.current_pipeline, actions)
            
            dim = [max(1, len(a)) for a in actions]
            if (registry.has_module(spreadsheet_pkg, 'CellLocation') and
                registry.has_module(spreadsheet_pkg, 'SheetReference')):
                modifiedPipelines = self.virtualCell.positionPipelines(
                    'PE#%d %s' % (QParameterExplorationTab.explorationId,
                                  self.controller.name),
                    dim[2], dim[1], dim[0], pipelines, self.controller)
            else:
                modifiedPipelines = pipelines

            mCount = []
            for p in modifiedPipelines:
                if len(mCount)==0:
                    mCount.append(0)
                else:
                    mCount.append(len(p.modules)+mCount[len(mCount)-1])
                
            # Now execute the pipelines
            totalProgress = sum([len(p.modules) for p in modifiedPipelines])
            progress = QtGui.QProgressDialog('Performing Parameter '
                                             'Exploration...',
                                             '&Cancel',
                                             0, totalProgress)
            progress.setWindowTitle('Parameter Exploration')
            progress.setWindowModality(QtCore.Qt.WindowModal)
            progress.show()

            QParameterExplorationTab.explorationId += 1
            interpreter = get_default_interpreter()
            for pi in xrange(len(modifiedPipelines)):
                progress.setValue(mCount[pi])
                QtCore.QCoreApplication.processEvents()
                if progress.wasCanceled():
                    break
                def moduleExecuted(objId):
                    if not progress.wasCanceled():
                        #progress.setValue(progress.value()+1)
                        #the call above was crashing when used by multithreaded
                        #code, replacing with the call below (thanks to Terence
                        #for submitting this fix). 
                        QtCore.QMetaObject.invokeMethod(progress, "setValue", 
                                        QtCore.Q_ARG(int,progress.value()+1))
                        QtCore.QCoreApplication.processEvents()
                kwargs = {'locator': self.controller.locator,
                          'current_version': self.controller.current_version,
                          'view': self.controller.current_pipeline_scene,
                          'module_executed_hook': [moduleExecuted],
                          'reason': 'Parameter Exploration',
                          'actions': performedActions[pi],
                          }
                interpreter.execute(modifiedPipelines[pi], **kwargs)
            progress.setValue(totalProgress)

    def exploreChange(self, notEmpty):
        """ exploreChange(notEmpty: bool) -> None
        echo the signal
        
        """
        self.emit(QtCore.SIGNAL('exploreChange(bool)'), notEmpty)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtGui, QtCore
from vistrails.core.modules.vistrails_module import ModuleError
import vistrails.core.system
import copy
import sys
import time
import os.path
import vistrails.gui.application
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

############################################################################

class QDebugger(QtGui.QWidget, QVistrailsPaletteInterface):
    """
    This class provides a dockable interface to the debugger tree.
    """
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent=parent)
        self.app = vistrails.gui.application.get_vistrails_application()
        self.inspector = QObjectInspector()
        layout = QtGui.QVBoxLayout()
        layout.setMargin(0)
        layout.setSpacing(0)
        layout.addWidget(self.inspector)
        self.setLayout(layout)
        # self.setTitleBarWidget(QtGui.QLabel("Debugger"))
        self.setWindowTitle("Debugger")
        self.controller = None
        self.vistrails_interpreter = get_default_interpreter()
        self.vistrails_interpreter.debugger = self

    def set_controller(self, c):
        """
        set_controller(c) -> None
        Set the current vistrails controller to be used by the debugger
        """
        self.controller = c
        self.update()

    def update_values(self):
        """
        update_vals() -> None
        Update the debugger after an execution with any values that become
        available on its input ports.
        """
        self.update(update_vals=True)
        
    def update(self, update_vals=False):
        """
        update(update_vals=False) -> None
        Update the debugger.  If the update requires querying modules for input
        changes, update_vals should be set to True
        """
        self.inspector.clear_modules()
        if self.controller is not None:
            pipeline = self.controller.current_pipeline
            if pipeline is None:
                return

            for module in pipeline.module_list:
                if module.is_breakpoint or module.is_watched:
                    self.inspector.add_module(module)
            if update_vals:
                (module_objects, _, _) = \
                    self.vistrails_interpreter.find_persistent_entities(
                        pipeline)
                for m_id in self.inspector.modules:
                    if (m_id in module_objects and 
                            module_objects[m_id] is not None):
                        self.inspector.update_values(m_id, module_objects[m_id])
                    elif module_objects[m_id] is None:
                        edges = pipeline.graph.edges_to(m_id)
                        self.inspector.update_inputs(m_id, module_objects, 
                                                     edges,
                                                      pipeline.connections)

    def closeEvent(self, e):
        """closeEvent(e) -> None
        Event handler called when the dialog is about to close."""
        self.emit(QtCore.SIGNAL("debuggerHidden()"))
                        
###############################################################################
#  QObjectInspector

class QObjectInspector(QtGui.QTreeWidget):
    """
    This class provides the ability to track and inspect breakpoints added to a pipeline.
    It is meant to be embedded in the QDebugger object to allow debugging of workflows in
    VisTrails
    """
    def __init__(self, parent=None):
        QtGui.QTreeWidget.__init__(self, parent)
        self.setColumnCount(2)
        self.modules = {}

    def clear_modules(self):
        """
        clear_modules() -> None
        Clear the current list of module breakpoints
        """
        self.modules = {}
        self.clear()
     
    def add_module(self, m):
        """
        add_module(m : core.vistrail.module.Module) -> None
        Add the give module, m, as a breakpoint.
        """
        # !!! This uses the core.vistrail.module.Module item
        item = QDebugModuleItem(self)
        item.setText(0, "%s (%d)" % (m.name, m.id))
        item.setText(1, "Module Type")
        self.modules[m.id] = item
#         self.add_dict(m, item)
#         self.add_ports(m, item, display_vals=get_vals)
        
    def update_values(self, m_id, persistent_module):
        """
        update_values(m_id: long, 
          persistent_module : subclass of core.modules.vistrails_module.Module)
        """
        module_item = self.modules[m_id]
        module_item.takeChildren()
        self.add_dict(persistent_module, module_item)
        self.add_ports(persistent_module, module_item, True)

    def update_inputs(self, m_id, persistent_map, edges, connections):
        input_ports = {}
        for upstream_id, c_id in edges:
            if upstream_id in persistent_map and \
                    persistent_map[upstream_id] is not None:
                persistent_module = persistent_map[upstream_id]
                connection = connections[c_id]
                try:
                    output_port = \
                        persistent_module.get_output(connection.source.name)
                    input_ports[connection.destination.name] = output_port
                except ModuleError:
                    input_ports[connection.destination.name] = None
        if len(input_ports) > 0:
            module_item = self.modules[m_id]
            module_item.takeChildren()
            inputs_item = QDebugModuleItem(module_item)
            inputs_item.setText(0, "inputPorts")
            inputs_item.setText(1, "")   
            for port_name, port_val in input_ports.iteritems():
                self.create_port_item(port_name, port_val, True, 
                                      inputs_item)
                
    def add_dict(self, m, parent_item):
        """
        add_dict(m, parent_item) -> None
        Add the dictionary associated with module m to be displayed 
        as part of the debug information for that breakpoint.
        """
        dict_item = QDebugModuleItem(parent_item)
        dict_item.setText(0, "__dict__")
        dict_item.setText(1, "")
        for k in m.__dict__.keys():
            d_val = QDebugModuleItem(dict_item)
            d_val.setText(0, str(k))
            d_val.setText(1, str(m.__dict__[k]))


    def create_port_item(self, port_name, port_value, display_vals=False,
                         parent=None):
        p_item = QDebugModuleItem(parent)
        p_item.setText(0, str(port_name))
        if display_vals:
            p_item.setText(1, str(port_value))
        else:
            typestr = str(port_value.__class__)
            typestr = typestr.split('.')
            typestr = typestr[len(typestr)-1]
            typestr = typestr[0:len(typestr)-2]
            p_item.setText(1, typestr)            
            
    def add_ports(self, m, parent_item, display_vals=False):
        """
        add_ports(m, item, display_vals=False) -> None
        Add port information from module m to the item being displayed in the debugger.
        If display_vals is True, fetch the appropriate values from the module's input ports.
        """
        inputs_item = QDebugModuleItem(parent_item)
        inputs_item.setText(0, "inputPorts")
        inputs_item.setText(1, "")
        for port_name in m.inputPorts:
            try:
                port_val = m.get_input_list(port_name)
                if len(port_val) == 1:
                    port_val = port_val[0]
            except ModuleError:
                port_val = None
            self.create_port_item(port_name, port_val, display_vals, 
                                  inputs_item)
        outputs_item = QDebugModuleItem(parent_item)
        outputs_item.setText(0, "outputPorts")
        outputs_item.setText(1, "")
        for port_name in m.outputPorts:
            try:
                port_val = m.get_output(port_name)
            except ModuleError:
                port_val = None
            self.create_port_item(port_name, port_val, display_vals, 
                                  outputs_item)

########################################################################
# QDebugModuleItem

class QDebugModuleItem(QtGui.QTreeWidgetItem):
    """
    This class provides a unique container for adding breakpoints in a workflow
    to the debugger.
    """
    def __init__(self, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent)
        



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.configuration import get_vistrails_configuration
""" This python module defines Connection class.
"""
import copy
from vistrails.db.domain import DBConnection
from vistrails.core.vistrail.port import PortEndPoint, Port

import unittest
from vistrails.db.domain import IdScope

################################################################################

class Connection(DBConnection):
    """ A Connection is a connection between two modules.
    Right now there's only Module connections.

    """

    ##########################################################################
    # Constructors and copy
    
    @staticmethod
    def from_port_specs(source, dest):
        """from_port_specs(source: PortSpec, dest: PortSpec) -> Connection
        Static method that creates a Connection given source and 
        destination ports.

        """
        conn = Connection()
        conn.source = copy.copy(source)
        conn.destination = copy.copy(dest)
        return conn
        
    @staticmethod
    def fromID(id):
        """fromTypeID(id: int) -> Connection
        Static method that creates a Connection given an id.

        """
        conn = Connection()
        conn.id = id
        conn.source.endPoint = PortEndPoint.Source
        conn.destination.endPoint = PortEndPoint.Destination
        return conn
    
    def __init__(self, *args, **kwargs):
        """__init__() -> Connection 
        Initializes source and destination ports.
        
        """
        DBConnection.__init__(self, *args, **kwargs)
        if self.id is None:
            self.db_id = -1
        if not len(self.ports) > 0:
            self.source = Port(type='source')
            self.destination = Port(type='destination')

    def __copy__(self):
        """__copy__() -> Connection -  Returns a clone of self.
        
        """
        return Connection.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBConnection.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Connection
        for port in cp.ports:
            Port.convert(port)
        return cp

    ##########################################################################

    @staticmethod
    def convert(_connection):
#        print "ports: %s" % _Connection._get_ports(_connection)
        if _connection.__class__ == Connection:
            return
        _connection.__class__ = Connection

        for port in _connection.ports:
            Port.convert(port)

    ##########################################################################
    # Properties

    id = DBConnection.db_id
    ports = DBConnection.db_ports
    
    def add_port(self, port):
        self.db_add_port(port)

    def _get_sourceId(self):
        """ _get_sourceId() -> int
        Returns the module id of source port. Do not use this function, 
        use sourceId property: c.sourceId 

        """
        return self.source.moduleId

    def _set_sourceId(self, id):
        """ _set_sourceId(id : int) -> None 
        Sets this connection source id. It updates both self.source.moduleId
        and self.source.id. Do not use this function, use sourceId
        property: c.sourceId = id

        """
        self.source.moduleId = id
        self.source.id = id
    sourceId = property(_get_sourceId, _set_sourceId)

    def _get_destinationId(self):
        """ _get_destinationId() -> int
        Returns the module id of dest port. Do not use this function, 
        use sourceId property: c.destinationId 

        """
        return self.destination.moduleId

    def _set_destinationId(self, id):
        """ _set_destinationId(id : int) -> None 
        Sets this connection destination id. It updates self.dest.moduleId.
        Do not use this function, use destinationId property: 
        c.destinationId = id

        """
        self.destination.moduleId = id
    destinationId = property(_get_destinationId, _set_destinationId)

    def _get_source(self):
        """_get_source() -> Port
        Returns source port. Do not use this function, use source property: 
        c.source 

        """
        try:
            return self.db_get_port_by_type('source')
        except KeyError:
            pass
        return None

    def _set_source(self, source):
        """_set_source(source: Port) -> None 
        Sets this connection source port. Do not use this function,
        use source property instead: c.source = source

        """
        try:
            port = self.db_get_port_by_type('source')
            self.db_delete_port(port)
        except KeyError:
            pass
        if source is not None:
            self.db_add_port(source)
    source = property(_get_source, _set_source)

    def _get_destination(self):
        """_get_destination() -> Port
        Returns destination port. Do not use this function, use destination
        property: c.destination 

        """
#        return self.db_ports['destination']
        try:
            return self.db_get_port_by_type('destination')
        except KeyError:
            pass
        return None

    def _set_destination(self, dest):
        """_set_destination(dest: Port) -> None
         Sets this connection destination port. Do not use this
        function, use destination property instead: c.destination = dest

        """
        try:
            port = self.db_get_port_by_type('destination')
            self.db_delete_port(port)
        except KeyError:
            pass
        if dest is not None:
            self.db_add_port(dest)
    destination = property(_get_destination, _set_destination)
    dest = property(_get_destination, _set_destination)

    ##########################################################################
    # Operators

    def __str__(self):
        """__str__() -> str - Returns a string representation of a Connection
        object. 

        """
        rep = "<connection id='%s'>%s%s</connection>"
        return  rep % (str(self.id), str(self.source), str(self.destination))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return (self.source == other.source and
                self.dest == other.dest)

    def equals_no_id(self, other):
        """Checks equality up to ids (connection and ports)."""
        if type(self) != type(other):
            return False
        return (self.source.equals_no_id(other.source) and
                self.dest.equals_no_id(other.dest))
    
################################################################################
# Testing


class TestConnection(unittest.TestCase):

    def create_connection(self, id_scope=IdScope()):
        from vistrails.core.vistrail.port import Port
        from vistrails.core.modules.basic_modules import identifier as basic_pkg
        source = Port(id=id_scope.getNewId(Port.vtType),
                      type='source', 
                      moduleId=21L, 
                      moduleName='String', 
                      name='value',
                      signature='(%s:String)' % basic_pkg)
        destination = Port(id=id_scope.getNewId(Port.vtType),
                           type='destination',
                           moduleId=20L,
                           moduleName='Float',
                           name='value',
                           signature='(%s:Float)' % basic_pkg)
        connection = Connection(id=id_scope.getNewId(Connection.vtType),
                                ports=[source, destination])
        return connection

    def test_copy(self):
        id_scope = IdScope()
        
        c1 = self.create_connection(id_scope)
        c2 = copy.copy(c1)
        self.assertEquals(c1, c2)
        self.assertEquals(c1.id, c2.id)
        c3 = c1.do_copy(True, id_scope, {})
        self.assertEquals(c1, c3)
        self.assertNotEquals(c1.id, c3.id)

    def test_serialization(self):
        import vistrails.core.db.io
        c1 = self.create_connection()
        xml_str = vistrails.core.db.io.serialize(c1)
        c2 = vistrails.core.db.io.unserialize(xml_str, Connection)
        self.assertEquals(c1, c2)
        self.assertEquals(c1.id, c2.id)

    def testEmptyConnection(self):
        """Tests sane initialization of empty connection"""
        c = Connection()
        self.assertEquals(c.source.endPoint, PortEndPoint.Source)
        self.assertEquals(c.destination.endPoint, PortEndPoint.Destination)
        
if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" contains parts from site.py for python2.7 needed when using py2app
"""

from __future__ import division

import sys

# for distutils.commands.install
# These values are initialized by the getuserbase() and getusersitepackages()
# functions, through the main() function when Python starts.
USER_SITE = None
USER_BASE = None

def getuserbase():
    """Returns the `user base` directory path.

    The `user base` directory can be used to store data. If the global
    variable ``USER_BASE`` is not initialized yet, this function will also set
    it.
    """
    global USER_BASE
    if USER_BASE is not None:
        return USER_BASE
    from sysconfig import get_config_var
    USER_BASE = get_config_var('userbase')
    return USER_BASE

def getusersitepackages():
    """Returns the user-specific site-packages directory path.

    If the global variable ``USER_SITE`` is not initialized yet, this
    function will also set it.
    """
    global USER_SITE
    user_base = getuserbase() # this will also set USER_BASE

    if USER_SITE is not None:
        return USER_SITE

    from sysconfig import get_path
    import os

    if sys.platform == 'darwin':
        from sysconfig import get_config_var
        if get_config_var('PYTHONFRAMEWORK'):
            USER_SITE = get_path('purelib', 'osx_framework_user')
            return USER_SITE

    USER_SITE = get_path('purelib', '%s_user' % os.name)
    return USER_SITE

class _Helper(object):
    """Define the builtin 'help'.
    This is a wrapper around pydoc.help (with a twist).
    
    """

    def __repr__(self):
        return "Type help() for interactive help, " \
               "or help(object) for help about object."
    def __call__(self, *args, **kwds):
        import pydoc
        return pydoc.help(*args, **kwds)


###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" Base classes for all Hadoop Modules """

from __future__ import division

from init import configuration
from vistrails.core.modules.basic_modules import File, String
from vistrails.gui.modules.python_source_configure import \
                                                PythonSourceConfigurationWidget
from vistrails.core.modules.config import ModuleSettings, IPort, OPort
from vistrails.core.modules.vistrails_module import Module, NotCacheable, \
                                                   ModuleError, ModuleSuspended
from remoteq.core.stack import select_machine, end_machine, use_machine, \
                                                                current_machine
from remoteq.batch.commandline import Subshell
from remoteq.batch.directories import CreateDirectory
import urllib
import xml.etree.cElementTree as ET

from init import RQModule

################################################################################
class HadoopBaseModule(RQModule):
    """
    The base class for all modules in the Hadoop package. We would
    like to implement a few basic functionality such as looking for
    Hadoop installation directory, execute Hadoop commands, etc. for
    subclasses to use
    
    """
    _settings = ModuleSettings(abstract=True)

    hadoop_configuration = None
    def __init__(self):
        Module.__init__(self)
    
    def get_hadoop_home(self, machine):
        HADOOP_HOME = machine.remote.send_command("echo $HADOOP_HOME").strip()
        if HADOOP_HOME == '':
            # raise ModuleError(self, 'HADOOP_HOME has to be defined')
            # it does not actually. This means it is part of the system.
            pass
        return HADOOP_HOME

    def read_site_config(self, machine):
        config = HadoopBaseModule.hadoop_configuration
        # For AWS
        core_site = config['home']+'/conf/core-site.xml'
        # for NYU/CUSP
        #core_site = config['home']+'/etc/hadoop/conf/core-site.xml'
        site_string = machine.remote.cat(core_site)
        root = ET.fromstring(site_string)
        for node in root:
            name = node.find('name').text
            value = node.find('value').text
            config[name] = value

    def get_hadoop_config(self, machine):
        if HadoopBaseModule.hadoop_configuration==None:
            hadoop_home = self.get_hadoop_home(machine)
            # paths to try in order
            streaming_paths = ['/share/hadoop/tools/lib/',   # AWS
                               '/usr/lib/hadoop-mapreduce/', # NYU/CUSP
                               '/contrib/streaming/']
            for path in streaming_paths:
                hs = hadoop_home + path
                command = ("python -c \"import os, os.path; print '' if not "
                           "os.path.exists('{0}') else ''.join([i for i in "
                           "os.listdir('{0}') if 'streaming' in i][-1:])\""
                           ).format(hs)
                streamingjar = machine.remote.send_command(command).strip()
                if streamingjar:
                    break
            if not streamingjar:
                raise ModuleError(self,
                                  'hadoop-streaming.jar not found. Please add '
                                  'its directory to list of supported paths.')
            hadoop = (hadoop_home + '/bin/hadoop') if hadoop_home else 'hadoop'
            hdfs = (hadoop_home + '/bin/hdfs') if hadoop_home else 'hdfs'
            if not machine.remote.command_exists(hdfs):
                hdfs = hadoop
            config = {'home': hadoop_home,
                      'hadoop': hadoop,
                      'hdfs': hdfs,
                      'streaming.jar': hs + streamingjar}
            HadoopBaseModule.hadoop_configuration = config
            # reading configuration files are error-prone
            #self.read_site_config(machine)
            config['fs.defaultFS'] = ''
            # can access config only if hdfs command exists
            if hadoop != hdfs:
                config['fs.defaultFS'] = \
                    self.call_hdfs('getconf -confKey fs.defaultFS', machine)
        return HadoopBaseModule.hadoop_configuration

    def add_prefix(self, path, machine): 
        aliases = []
        if configuration.check('uris') and configuration.uris:
            aliases = dict([(uri.split('#')[1], uri.split('#')[0])
                            for uri in configuration.uris.split(';')])
        if path in aliases:
            return aliases[path]
        if configuration.check('defaultFS'):
            prefix = configuration.defaultFS
            return prefix + path
        else:
            prefix = self.get_hadoop_config(machine)['fs.defaultFS']
            return prefix + '/' + path


    def call_hadoop(self, arguments, workdir, identifier, machine):
        self.is_cacheable = lambda *args, **kwargs: False
        config = self.get_hadoop_config(machine)
        argList = [config['hadoop']]
        if type(arguments) in [str, unicode]:
            argList += arguments.split(' ')
        elif type(arguments)==list:
            argList += arguments
        else:
            raise ModuleError(self, 'Invalid argument types to hadoop')

        # 1. this version returns when finished
        #return subprocess.call(argList)
        # 2. this version reads the results incrementally
#         expect = machine.remote._expect_token
#         machine.remote.push_expect(None) # Do not wait for call to finish
#         result =  machine.remote.send_command(" ".join(argList)).strip()
#         machine.remote.pop_expect() # restore expect
#         # We could show the output in a gui
#         print "**** hadoop streaming running ****"
#         print result,
#         while not expect in result:
#             output = machine.remote.consume_output()
#             if output:
#                 print output,
#             result += output
        # 3. The final version should detach the process on the server
        use_machine(machine)
        cdir = CreateDirectory("remote", workdir)
        job = Subshell("remote", command=" ".join(argList),
                       working_directory=workdir, identifier=identifier,
                       dependencies=[cdir])
        job.run()
        finished = job.finished()
        if not finished:
            status = job.status()
            # The Subshell class provides the JobHandle interface, i.e.
            # finished()
            raise ModuleSuspended(self, '%s' % status, handle=job)
        self.is_cacheable = lambda *args, **kwargs: True
        return job.standard_error()

    def call_hdfs(self, arguments, machine):
        config = self.get_hadoop_config(machine)
        argList = [config['hdfs']]
        if type(arguments) in [str, unicode]:
            argList += arguments.split(' ')
        elif type(arguments)==list:
            argList += arguments
        else:
            raise ModuleError(self, 'Invalid argument types to hdfs: %s'%type(arguments))

        result =  machine.remote.send_command(" ".join(argList)).strip()
        return result
        #return subprocess.call(argList)

################################################################################
class PythonSourceToFileConfigurationWidget(PythonSourceConfigurationWidget):
    """
    A simple python source configuration widget, i.e. hiding all of
    the input/output port table
    
    """
    def __init__(self, module, controller, parent=None):
        PythonSourceConfigurationWidget.__init__(self, module,
                                                 controller, parent)
        self.inputPortTable.hide()
        self.outputPortTable.hide()
        self.setWindowTitle('Python Source Editor')

################################################################################
class PythonSourceToFile(Module):
    """
    This is the class for specifying a python code snippet for running
    with Hadoop Streaming, it will take its contents and output to a
    temporary Python file. The code will not be passed around.
    
    """
    _settings = ModuleSettings(namespace='hadoop',
                        configure_widget=PythonSourceToFileConfigurationWidget)

    _input_ports = [IPort('Input File', File),
                    IPort('source', String, optional=True)]

    _output_ports = [OPort('Temporary File', File)]

    def compute(self):
        inputFile = self.force_get_input('Input File')

        if inputFile!=None:
#            tempFile = file_pool.make_local_copy(inputFile.name)
            tempFile = inputFile
        else:
            source = urllib.unquote(self.force_get_input('source', ''))
            tempFile = self.interpreter.filePool.create_file()
            f = open(tempFile.name, 'w')
            f.write(source)
            f.close()
        self.set_output('Temporary File', tempFile)
            

################################################################################
def register():
    return [HadoopBaseModule, PythonSourceToFile]
    
